from geometry import Positionable
from logger import Logger
from trackers import create_tracker


class Widget(Positionable, Logger):
    """
    A widget represents a 2D object on the screen in which gestures can occur.
    Handlers for a specific gesture type can be bound to a widget. The widget
    will
    """
    def __init__(self, x=None, y=None):
        Positionable.__init__(self, x, y)

        # Map of gesture types to gesture trackers
        self.trackers = {}

        # Map of gesture types to a list of handlers for that type
        self.handlers = {}

        # Widget tree references
        self.parent = None
        self.children = []

    def add_widget(self, widget):
        """
        Add a new child widget.
        """
        self.children.append(widget)
        widget.set_parent(self)

    def remove_widget(self, widget):
        """
        Remove a child widget.
        """
        self.children.remove(widget)
        widget.set_parent(None)

    def set_parent(self, widget):
        """
        Set a new parent widget. If a parent widget has already been assigned,
        remove the widget from that parent first.
        """
        if self.parent:
            self.parent.remove_widget(self)

        self.parent = widget

    def unbind(self, gesture_type, handler=None):
        """
        Unbind a single handler, or all handlers bound to a gesture type.
        Remove the corresponding tracker if it is no longer needed.
        """
        if gesture_type not in self.handlers:
            raise KeyError('Gesture type "%s" is not bound.')

        if handler:
            # Remove a specific handler
            self.handlers[gesture_type].remove(handler)

            # Only remove the handler list and optionally the tracker if no
            # other handlers exist for the gesture type
            if self.handlers[gesture_type]:
                return
        else:
            # Remove handler list
            del self.handlers[gesture_type][:]

        # Check if any other handlers need the tracker
        for gtype in self.trackers[gesture_type].gesture_types:
            if gtype in self.handlers:
                return

        # No more handlers are bound, remove unused tracker and handlers
        del self.trackers[gesture_type]

        for gtype in self.trackers[gesture_type].gesture_types:
            del self.handlers[gtype]

    def bind(self, gesture_type, handler, remove_existing=False):
        """
        Bind a handler to the specified type of gesture. Create a tracker for
        the gesture type if it does not exists yet.
        """
        if gesture_type not in self.handlers:
            tracker = create_tracker(gesture_type, self)
            self.trackers[gesture_type] = tracker
            self.handlers[gesture_type] = []

            # Create empty tracker lists for all supported gestures
            for gtype in tracker.gesture_types:
                self.handlers[gtype] = []
        elif remove_existing:
            del self.handlers[gesture_type][:]

        self.handlers[gesture_type].append(handler)

    def __getattr__(self, name):
        """
        Allow calls like:
        widget.on_gesture(...)
        instead of:
        widget.bind('gesture', ...)
        """
        if len(name) < 4 or name[:3] != 'on_':
            raise AttributeError("'%s' has no attribute '%s'"
                                 % (self.__class__.__name__, name))

        return lambda handler: self.bind(name[3:], handler)

    def contains_event(self, event):
        """
        Check if the coordinates of an event are contained within this widget.
        """
        raise NotImplementedError

    def handle_event(self, event):
        """
        Delegate a triggered event to gesture trackers and child widgets.  A
        handler can stop the delegation of the event, preventing it from being
        delegated to child widgets.
        """
        for tracker in set(self.trackers.itervalues()):
            tracker.handle_event(event)

        if not event.is_delegation_stopped():
            for child in self.children:
                if child.contains_event(event):
                    child.handle_event(event)

    def handle_gesture(self, gesture):
        """
        Handle a gesture that is triggered by a gesture tracker. First, all
        handlers bound to the gesture type are called. Second, if the gesture's
        propagation has not been stopped by a handler, the gesture is
        propagated to the parent widget.
        """
        for handler in self.handlers.get(gesture.get_type(), ()):
            handler(gesture)

        if self.parent and not gesture.is_propagation_stopped():
            self.parent.handle_gesture(gesture)
